#ifndef PARSER_H
#define PARSER_H

#include "config.h"

#include <list>
#include <vector>
using namespace std;

class Parser;

class Node
	{
public:
	bool rhoterm;
	int ndHole;
	virtual void print(std::ostream& o);
	virtual void checkSoundnessAndCalcul();
	virtual bool containVar(int idvar);
	virtual void replace(int idvar,Node *n);
	
	/**
	* duplicate the node
	*/
	virtual Node* duplicate() = 0;
    virtual ~Node();
	virtual bool isEqual(Node *n);
	list<unsigned int> *holePosition;
	Node();
	static int ndterms;
	virtual Node *find(list<unsigned int> &path);
	virtual Node* buildOneAndReplace(list<unsigned int> &path,Node *sub);
	};

class Hole : public Node
	{
public:
	void print(std::ostream& o);
	void checkSoundnessAndCalcul();
	Node* duplicate();
	Hole();
	~Hole();
	};
	
class Clash : public Node
	{
	void print(std::ostream& o);
	Node* duplicate();
	};
	
class Trivial : public Node
	{
	void print(std::ostream& o);
	Node* duplicate();
	};

/* for the integer exponents that are behind a numerical variable */
class NumericalVar
	{
	Parser *parser;
public:
	int id;
	NumericalVar(const char* name,Parser *parser,bool automatic=true);
	void print(std::ostream& o);
	NumericalVar* duplicate();
	bool isEqual(Node *n);
	};
	
class Variable : public Node
	{
	Parser *parser;
public:
	int id;
	Variable(const char *name,Parser *parser);
	void print(std::ostream& o);
	void checkSoundnessAndCalcul();
	bool containVar(int idvar);
	Node* duplicate();
	bool isEqual(Node *n);
	};
	
class Function : public Node
	{
	Parser *parser;
public:
	int id;
	vector<Node*> arguments;
	Function(const char* name,Parser *parser);
	Function(int id,Parser *parser);
	Function();
	void addArgument(Node *n);
	void print(std::ostream& o);
	void checkSoundnessAndCalcul();
	bool containVar(int idvar);
	void replace(int idvar,Node *n);
	void setName(const char* name,Parser *parser);
	Node* buildOneAndReplace(list<unsigned int> &path,Node *sub);
	Node* duplicate();
	bool isEqual(Node *n);
	Node *find(list<unsigned int> &path);
	};

class Or : public Node
	{
public:
	list<Node*> branchs;
	Or();
	void addBranch(Node *n);
	void print(std::ostream& o);
	void checkSoundnessAndCalcul();
	bool containVar(int idvar);
	void replace(int idvar,Node *n);
	Node* duplicate();
	bool isEqual(Node *n);
	};
	
class Condition
	{
public:
	NumericalVar* n1;
	virtual void print(std::ostream& o);
    virtual ~Condition();
	virtual Condition* duplicate() = 0;
	virtual bool isEqual(Condition *n) = 0;
	};

//n1 = a
class CondA : public Condition
	{
public:
	int a;
	CondA(NumericalVar* n1,int a);
	void print(std::ostream& o);
	Condition* duplicate();
	bool isEqual(Condition *n);
	};

//n1 = a * n2 + b
class CondANplusB : public Condition
	{
public:
	int a;
	NumericalVar* n2;
	int b;
	CondANplusB(NumericalVar* n1,int a,NumericalVar* n2,int b);
	void print(std::ostream& o);
	Condition* duplicate();
	bool isEqual(Condition *n);
	};
	
//n1 = n2 + n3
class CondNplusN : public Condition
	{
public:
	NumericalVar* n2;
	NumericalVar* n3;
	CondNplusN(NumericalVar* n1,NumericalVar* n2,NumericalVar* n3);
	void print(std::ostream& o);
	Condition* duplicate();
	bool isEqual(Condition *n);
	};
	
//n1 = n2
class CondN : public Condition
	{
public:
	NumericalVar* n2;
	CondN(NumericalVar* n1,NumericalVar* n2);
	void print(std::ostream& o);
	Condition* duplicate();
	bool isEqual(Condition *n);
	};
	
	
class And : public Node
	{
public:
	list<Node*> branchs;
	list<Condition*> conditions;
	And();
	void addBranch(Node *n);
	void addCondition(Condition *c);
	void print(std::ostream& o);
	void printNoBracket(std::ostream& o);
	void checkSoundnessAndCalcul();
	bool containVar(int idvar);
	void replace(int idvar,Node *n);
	Node* duplicate();
	bool isEqual(Node *n);
	};

class Equal : public Node
	{
public:
	Node *n1;
	Node *n2;
	Equal(Node *n1,Node *n2);
	void print(std::ostream& o); 
	void checkSoundnessAndCalcul();
	bool containVar(int idvar);
	void replace(int idvar,Node *n);
	Node* duplicate();
	bool isEqual(Node *n);
	};

class Expo : public Node
	{
public:
	bool free;
	Node *n1;
	Node *n2;
	union
		{
		int expNum;
		NumericalVar *var;
		};
	Expo(Node *n1,int num,Node *n2);
	Expo(Node *n1,NumericalVar *num,Node *n2);
	void print(std::ostream& o);
	void checkSoundnessAndCalcul();
	bool containVar(int idvar);
	void replace(int idvar,Node *n);
	Node* buildOne(Node *sub = NULL);
	Node* buildTwo(Node *sub = NULL);
	Node* build(int num,Node *sub = NULL);
	Node* duplicate();
	bool isEqual(Node *n);
	};
	
class FonctionTabElement
	{
public:
	string name;
	int arity;
	};
	
class VariableTabElement
	{
public:
	string name;
	};
	
class NumericalVariableTabElement
	{
public:
	string name;
	bool automatic;
	};

class StackElement
	{
public:
	list<unsigned int> position;
	Node* n;
	Node **direct;
	StackElement(Node *n,Node **direct);
	StackElement(Node *n,Node **direct,list<unsigned int> &parentpos,int me);
	StackElement();
	};

class NodeIterator
	{
public:
	enum Mode
		{
		DEPTH_FIRST,
		BREADTH_FIRST
		};
private:
	Mode mode;
	list<StackElement> stack;
	StackElement last;
	Node *n;
	Node **direct;
	bool allowUnderEqual;
public:
	void setUnderEqual(bool u);
	NodeIterator(Node *n,Node **direct,Mode mode);
	const list<unsigned int> &getPosition();
	Node* next();
	void reset();
	void replaceBy(Node *n);
	};

class Parser
	{
	int autoNumericalValCur;
	int autoContstantValCur;
public:
	vector<FonctionTabElement> fonctiontab;
	vector<VariableTabElement> variabletab;
	vector<NumericalVariableTabElement> numericalvariabletab;
	Node *result;
	Configuration *config;
	
	void checkSoundness();
	Parser(Configuration *config);
	string genNumericalValName();
	string genContstantValName();
	};

#endif

